#include <ftl/filters/smoothing.hpp>
#include "smoothing_cuda.hpp"

#include <ftl/cuda/normals.hpp>

using ftl::filters::DepthSmoother;
using ftl::filters::MLSSmoother;
using ftl::codecs::Channel;
using cv::cuda::GpuMat;

DepthSmoother::DepthSmoother(nlohmann::json &config) : ftl::Filter(config) {

}

DepthSmoother::~DepthSmoother() {

}

void DepthSmoother::filter(ftl::rgbd::Frame &f, ftl::rgbd::Source *s, cudaStream_t stream) {
    float var_thresh = value("variance_threshold", 0.0002f);
    //bool do_smooth = value("pre_smooth", false);
    int levels = max(0, min(value("levels",0), 4));
    int iters = value("iterations",5);

    //if (!do_smooth) return;

    for (int i=0; i<iters; ++i) {
        ftl::cuda::smoothing_factor(
            f.createTexture<float>(Channel::Depth),
            f.createTexture<float>(Channel::Energy, ftl::rgbd::Format<float>(f.get<cv::cuda::GpuMat>(Channel::Depth).size())),
            f.createTexture<float>(Channel::Smoothing, ftl::rgbd::Format<float>(f.get<cv::cuda::GpuMat>(Channel::Depth).size())),
            var_thresh,
            s->parameters(), 0
        );
    }

    LOG(INFO) << "PARAMS DEPTHS  " << s->parameters().minDepth << "," << s->parameters().maxDepth;

    for (int i=0; i<levels; ++i) {
        var_thresh *= 2.0f;
        auto &dmat = f.get<GpuMat>(Channel::Depth); 
        cv::cuda::resize(dmat, frames_[i].create<GpuMat>(Channel::Depth), cv::Size(dmat.cols / (2*(i+1)), dmat.rows / (2*(i+1))), 0.0, 0.0, cv::INTER_NEAREST);

        ftl::cuda::smoothing_factor(
            frames_[i].createTexture<float>(Channel::Depth),
            frames_[i].createTexture<float>(Channel::Energy, ftl::rgbd::Format<float>(frames_[i].get<GpuMat>(Channel::Depth).size())),
            frames_[i].createTexture<float>(Channel::Smoothing, ftl::rgbd::Format<float>(frames_[i].get<GpuMat>(Channel::Depth).size())),
            var_thresh,
            s->parameters(), 0
        );

        cv::cuda::resize(frames_[i].get<GpuMat>(Channel::Smoothing), temp_, f.get<cv::cuda::GpuMat>(Channel::Depth).size(), 0.0, 0.0, cv::INTER_LINEAR);
        cv::cuda::add(temp_, f.get<GpuMat>(Channel::Smoothing), f.get<GpuMat>(Channel::Smoothing));
    }

    //cv::cuda::subtract(f.get<GpuMat>(Channel::Depth), f.get<GpuMat>(Channel::Smoothing), f.get<GpuMat>(Channel::Depth));
}


// ===== MLS ===================================================================

MLSSmoother::MLSSmoother(nlohmann::json &config) : ftl::Filter(config) {

}

MLSSmoother::~MLSSmoother() {

}

void MLSSmoother::filter(ftl::rgbd::Frame &f, ftl::rgbd::Source *s, cudaStream_t stream) {
	//bool do_smooth = value("mls_smooth", false);
	float thresh = value("mls_threshold", 0.04f);
	int iters = value("mls_iterations", 1);
	int radius = value("mls_radius",2);

	//if (!do_smooth) return;

	if (!f.hasChannel(Channel::Normals)) {
		ftl::cuda::normals(
			f.createTexture<float4>(Channel::Normals, ftl::rgbd::Format<float4>(f.get<cv::cuda::GpuMat>(Channel::Depth).size())),
			f.createTexture<float>(Channel::Depth),
			s->parameters(), 0
		);
	}

	for (int i=0; i<iters; ++i) {
		ftl::cuda::mls_smooth(
			f.createTexture<float4>(Channel::Normals),
			f.createTexture<float4>(Channel::Points, ftl::rgbd::Format<float4>(f.get<cv::cuda::GpuMat>(Channel::Depth).size())),
			f.createTexture<float>(Channel::Depth),
			f.createTexture<float>(Channel::Depth2, ftl::rgbd::Format<float>(f.get<cv::cuda::GpuMat>(Channel::Depth).size())),
			thresh,
			radius,
			s->parameters(),
			0
		);

		f.swapChannels(Channel::Depth, Channel::Depth2);
		f.swapChannels(Channel::Normals, Channel::Points);
	}
}

